@@ -1,21 +1,7 @@
 /*
- * Copyright 2019 Fabian Vogt <fabian@ritter-vogt.de>
- *
- * This program is free software; you can redistribute it and/or 
- * modify it under the terms of the GNU General Public License as
- * published by the Free Software Foundation; either version 3 of
- * the License or any later version accepted by the membership of
- * KDE e.V. (or its successor approved by the membership of KDE
- * e.V.), which shall act as a proxy defined in Section 14 of
- * version 3 of the license.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   SPDX-FileCopyrightText: 2019-2020 Fabian Vogt <fabian@ritter-vogt.de>
+   SPDX-FileCopyrightText: 2019-2020 Alexander Saoutkin <a.saoutkin@gmail.com>
+   SPDX-License-Identifier: GPL-3.0-or-later
 */
 
 #include <fcntl.h>
@@ -32,20 +18,18 @@
 #include <QtDBus/QDBusReply>
 #include <QDebug>
 #include "kiofuse_interface.h"
+#include "kiofuseprivate_interface.h"
 
 class FileOpsTest : public QObject
 {
 	Q_OBJECT
-public:
-	FileOpsTest() : m_kiofuse_iface(QStringLiteral("org.kde.KIOFuse"),
-	                 QStringLiteral("/org/kde/KIOFuse"),
-	                 QDBusConnection::sessionBus()) {}
 
 private Q_SLOTS:
 	void initTestCase();
 	void cleanupTestCase();
 
 	void testDBusErrorReply();
+	void testLocalPathToRemoteUrl();
 	void testLocalFileOps();
 	void testLocalDirOps();
 	void testCreationOps();
@@ -55,18 +39,27 @@ private Q_SLOTS:
 	void testKioErrorMapping();
 	void testRootLookup();
 	void testFilenameEscaping();
+	void testDirRefresh();
+	void testFileRefresh();
+	void testSymlinkRefresh();
+	void testTypeRefresh();
 #ifdef WASTE_DISK_SPACE
 	void testReadWrite4GBFile();
 #endif // WASTE_DISK_SPACE
 
 private:
 	QDateTime roundDownToSecond(QDateTime dt);
+	bool forceNodeTimeout();
 
-	org::kde::KIOFuse::VFS m_kiofuse_iface;
+	org::kde::KIOFuse::VFS m_kiofuse_iface{QStringLiteral("org.kde.KIOFuse"),
+		                                   QStringLiteral("/org/kde/KIOFuse"),
+		                                   QDBusConnection::sessionBus()};
+	org::kde::KIOFuse::Private m_kiofuseprivate_iface{QStringLiteral("org.kde.KIOFuse"),
+		                                              QStringLiteral("/org/kde/KIOFuse"),
+		                                              QDBusConnection::sessionBus()};
 	QTemporaryDir m_mountDir;
 };
 
-
 void FileOpsTest::initTestCase()
 {
 	// QTemporaryDir would otherwise rm -rf on destruction,
@@ -76,7 +69,11 @@ void FileOpsTest::initTestCase()
 
 	QProcess kiofuseProcess;
 	kiofuseProcess.setProgram(programpath);
+#ifdef TEST_CACHE_BASED_IO
+	kiofuseProcess.setArguments(QStringList() << m_mountDir.path() << QStringLiteral("--disable-filejob-io"));
+#else
 	kiofuseProcess.setArguments({m_mountDir.path()});
+#endif
 	kiofuseProcess.setProcessChannelMode(QProcess::ForwardedChannels);
 
 	kiofuseProcess.start();
@@ -116,6 +113,30 @@ void FileOpsTest::testDBusErrorReply()
 	QCOMPARE(reply.error().name(), QStringLiteral("org.kde.KIOFuse.VFS.Error.SchemeNotSupported"));
 }
 
+void FileOpsTest::testLocalPathToRemoteUrl()
+{
+	QDBusPendingReply<QString> errorReply;
+	// mtp:/ -> Remote URL can't possibly be location of KIOFuse mount.
+	// / -> Root can't possibly be location of KIOFuse mount.
+	// m_mountDir -> Whilst this is in the KIOFuse mount, no remote URL exists for it
+	for(auto url : {QStringLiteral("mtp:/"), QStringLiteral("/"), m_mountDir.path()})
+	{
+		errorReply = m_kiofuse_iface.remoteUrl(url);
+		errorReply.waitForFinished();
+		QVERIFY2(errorReply.isError(), qPrintable(url));
+		QCOMPARE(errorReply.error().name(), QStringLiteral("org.kde.KIOFuse.VFS.Error.RemoteURLNotFound"));
+	}
+
+	QTemporaryFile localFile;
+	QVERIFY(localFile.open());
+	localFile.close(); // Force creation of file to avoid empty fileName()
+	QString remoteUrl = QStringLiteral("file://%1").arg(localFile.fileName());
+	QString reply = m_kiofuse_iface.mountUrl(remoteUrl).value();
+	QVERIFY(!reply.isEmpty());
+	QString calculatedRemoteUrl = m_kiofuse_iface.remoteUrl(reply).value();
+	QCOMPARE(remoteUrl, calculatedRemoteUrl);
+}
+
 void FileOpsTest::testLocalFileOps()
 {
 	QTemporaryFile localFile;
@@ -150,7 +171,6 @@ void FileOpsTest::testLocalFileOps()
 	QCOMPARE(mirroredFileInfo.lastRead(), roundDownToSecond(localFileInfo.lastRead()));
 
 	QVERIFY(mirroredFile.open(QIODevice::ReadWrite));
-
 	// Test touching the file
 	struct timespec times[2] = {{time_t(localFileInfo.lastModified().toSecsSinceEpoch()) + 42, 0},
 	                            {time_t(localFileInfo.lastRead().toSecsSinceEpoch()) + 1, 0}};
@@ -463,17 +483,21 @@ void FileOpsTest::testRenameOps()
 	                   RENAME_NOREPLACE), -1);
 	QCOMPARE(errno, EEXIST);
 #endif
+
 	QCOMPARE(rename(mirrorDir.filePath(QStringLiteral("dirb/fileb")).toUtf8().data(),
 	                mirrorDir.filePath(QStringLiteral("dirb/filec")).toUtf8().data()), 0);
-
 	QVERIFY(!QFile::exists(localDir.filePath(QStringLiteral("dirb/fileb"))));
 	QVERIFY(QFile::exists(localDir.filePath(QStringLiteral("dirb/filec"))));
 	QVERIFY(!QFile::exists(mirrorDir.filePath(QStringLiteral("dirb/fileb"))));
 	QVERIFY(QFile::exists(mirrorDir.filePath(QStringLiteral("dirb/filec"))));
 
-	// Both handles must still be valid
 	QVERIFY(overwrittenFile.seek(0));
+#ifdef TEST_CACHE_BASED_IO
+	// Both handles must still be valid
 	QCOMPARE(overwrittenFile.readAll(), QStringLiteral("data").toUtf8());
+#else
+	// Doesn't apply to FileJob (KIO::open) I/O
+#endif
 
 	localFile.close();
 	localFile.setFileName(localDir.filePath(QStringLiteral("dirb/filec")));
@@ -513,6 +537,7 @@ void FileOpsTest::testDeletionOps()
 	QCOMPARE(rmdir(dir.path().toUtf8().data()), -1);
 	QCOMPARE(errno, ENOTEMPTY);
 
+#ifdef TEST_CACHE_BASED_IO
 	// Delete the file
 	QCOMPARE(rmdir(file.fileName().toUtf8().data()), -1);
 	QCOMPARE(errno, ENOTDIR);
@@ -532,6 +557,17 @@ void FileOpsTest::testDeletionOps()
 	// Make sure the file is still open
 	QVERIFY(file.seek(0));
 	QCOMPARE(file.readAll(), QStringLiteral("someweirdstring").toUtf8());
+#else
+	// FileJob-based nodes only unlink if the file isn't open
+	QCOMPARE(rmdir(file.fileName().toUtf8().data()), -1);
+	QCOMPARE(errno, ENOTDIR);
+	QCOMPARE(unlink(file.fileName().toUtf8().data()), -1);
+	file.close();
+	QCOMPARE(unlink(file.fileName().toUtf8().data()), 0);
+	QVERIFY(!file.exists());
+	QVERIFY(!QFile::exists(localDir.filePath(QStringLiteral("dir/file"))));
+#endif
+
 
 	// Not implemented: Link the file back into the tree, if possible
 	// QCOMPARE(link(QStringLiteral("/proc/self/fd/%1").arg(file.handle()).toUtf8().data(),
@@ -627,6 +663,152 @@ void FileOpsTest::testFilenameEscaping()
 	}
 }
 
+void FileOpsTest::testDirRefresh()
+{
+	QTemporaryDir localDir;
+	QVERIFY(localDir.isValid());
+
+	// Mount the temporary dir
+	QString reply = m_kiofuse_iface.mountUrl(QStringLiteral("file://%1").arg(localDir.path())).value();
+	QVERIFY(!reply.isEmpty());
+
+	QDir mirrorDir(reply);
+	QVERIFY(mirrorDir.exists());
+
+	// readdir must not have any content yet
+	QCOMPARE(mirrorDir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot).count(), 0);
+
+	QFile newFile(localDir.filePath(QStringLiteral("newFile")));
+	QVERIFY(newFile.open(QFile::ReadWrite));
+
+	// Verify that the file is part of a dirlist after refresh
+	QCOMPARE(mirrorDir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot).count(), 0);
+	QVERIFY(forceNodeTimeout());
+	QCOMPARE(mirrorDir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot),
+	         QStringList{QStringLiteral("newFile")});
+
+	// Delete the file
+	newFile.close();
+	QVERIFY(newFile.remove());
+
+	// Verify that it disappears from the dirlist after refresh
+	QCOMPARE(mirrorDir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot), QStringList{QStringLiteral("newFile")});
+	QVERIFY(forceNodeTimeout());
+	QCOMPARE(mirrorDir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot).count(), 0);
+
+	// Recreate the file
+	QVERIFY(newFile.open(QFile::ReadWrite));
+
+	// Verify that access is immediately possible again (lookup is "optimistic")
+	QVERIFY(QFile::exists(mirrorDir.filePath(QStringLiteral("newFile"))));
+
+	// Delete the file again
+	newFile.close();
+	QVERIFY(newFile.remove());
+
+	// Verify that after a refresh it's dropped
+	QVERIFY(forceNodeTimeout());
+	QVERIFY(!QFile::exists(mirrorDir.filePath(QStringLiteral("newFile"))));
+}
+
+void FileOpsTest::testFileRefresh()
+{
+	QTemporaryDir localDir;
+	QVERIFY(localDir.isValid());
+
+	// Mount the temporary dir
+	QString reply = m_kiofuse_iface.mountUrl(QStringLiteral("file://%1").arg(localDir.path())).value();
+	QVERIFY(!reply.isEmpty());
+
+	QDir mirrorDir(reply);
+	QVERIFY(mirrorDir.exists());
+
+	// readdir must not have any content yet
+	QCOMPARE(mirrorDir.entryList(QDir::AllEntries | QDir::NoDotAndDotDot).count(), 0);
+
+	QFile localFile(localDir.filePath(QStringLiteral("newFile")));
+	QVERIFY(localFile.open(QFile::ReadWrite));
+
+	QFile mirrorFile(mirrorDir.filePath(QStringLiteral("newFile")));
+	QVERIFY(mirrorFile.open(QFile::ReadOnly));
+	QCOMPARE(mirrorFile.size(), 0); // File is empty
+	QCOMPARE(mirrorFile.readAll(), QByteArray{});
+	QVERIFY(mirrorFile.permissions() & QFile::ReadOther); // Has default perms
+
+	QCOMPARE(localFile.write("teststring", 10), 10); // Write some data
+	QVERIFY(localFile.flush());
+	QVERIFY(localFile.setPermissions(localFile.permissions() & ~QFile::ReadOther)); // Change perms
+	QCOMPARE(mirrorFile.size(), 0); // File is still empty
+	QVERIFY(forceNodeTimeout());
+
+	// Without reopening, it has the new content and perms now
+	QCOMPARE(mirrorFile.size(), 10);
+	QCOMPARE(mirrorFile.readAll(), QStringLiteral("teststring").toUtf8());
+	QCOMPARE(mirrorFile.permissions() & QFile::ReadOther, 0); // Has changed perms
+}
+
+void FileOpsTest::testSymlinkRefresh()
+{
+	QTemporaryDir localDir;
+	QVERIFY(localDir.isValid());
+
+	// Mount the temporary dir
+	QString reply = m_kiofuse_iface.mountUrl(QStringLiteral("file://%1").arg(localDir.path())).value();
+	QVERIFY(!reply.isEmpty());
+
+	QDir mirrorDir(reply);
+	QVERIFY(mirrorDir.exists());
+
+	// Create a symlink
+	QCOMPARE(symlink("/oldtarget", localDir.filePath(QStringLiteral("symlink")).toUtf8().data()), 0);
+	QCOMPARE(QFile::symLinkTarget(mirrorDir.filePath(QStringLiteral("symlink"))), QStringLiteral("/oldtarget"));
+
+	// Change the symlink
+	QVERIFY(QFile::remove(localDir.filePath((QStringLiteral("symlink")))));
+	QCOMPARE(symlink("/newtarget", localDir.filePath(QStringLiteral("symlink")).toUtf8().data()), 0);
+
+	QVERIFY(forceNodeTimeout());
+
+	QCOMPARE(QFile::symLinkTarget(mirrorDir.filePath(QStringLiteral("symlink"))), QStringLiteral("/newtarget"));
+}
+
+void FileOpsTest::testTypeRefresh()
+{
+	QTemporaryDir localDir;
+	QVERIFY(localDir.isValid());
+
+	// Mount the temporary dir
+	QString reply = m_kiofuse_iface.mountUrl(QStringLiteral("file://%1").arg(localDir.path())).value();
+	QVERIFY(!reply.isEmpty());
+
+	QDir mirrorDir(reply);
+	QVERIFY(mirrorDir.exists());
+
+	// Create a file and directory
+	QFile localFile(localDir.filePath(QStringLiteral("changingtodir")));
+	QVERIFY(localFile.open(QFile::ReadWrite));
+	QVERIFY(QDir(localDir.path()).mkdir(QStringLiteral("changingtofile")));
+
+	// Open it on the mirror
+	QFile changingMirrorFile(mirrorDir.filePath(QStringLiteral("changingtodir")));
+	QVERIFY(changingMirrorFile.open(QFile::ReadOnly));
+
+	// Replace the file locally  with a directory
+	QVERIFY(localFile.remove());
+	QVERIFY(QDir(localDir.path()).mkdir(QStringLiteral("changingtodir")));
+
+	QVERIFY(forceNodeTimeout());
+
+	// Verify that it's a directory now
+	struct stat st;
+	QCOMPARE(stat(qPrintable(changingMirrorFile.fileName()), &st), 0);
+	QCOMPARE(st.st_mode & S_IFMT, S_IFDIR);
+
+	// The opened file still refers to the (now deleted) file
+	QCOMPARE(fstat(changingMirrorFile.handle(), &st), 0);
+	QCOMPARE(st.st_mode & S_IFMT, S_IFREG);
+}
+
 #ifdef WASTE_DISK_SPACE
 void FileOpsTest::testReadWrite4GBFile()
 {
@@ -634,7 +816,7 @@ void FileOpsTest::testReadWrite4GBFile()
 	QVERIFY(localFile.open());
 
 	// Mount the temporary file
-	QDBusPendingReply reply = m_kiofuse_iface.mountUrl(QStringLiteral("file://%1").arg(localFile.fileName())).value();
+	QString reply = m_kiofuse_iface.mountUrl(QStringLiteral("file://%1").arg(localFile.fileName())).value();
 	QVERIFY(!reply.isEmpty());
 
 	QFile mirroredFile(reply);
@@ -670,6 +852,13 @@ QDateTime FileOpsTest::roundDownToSecond(QDateTime dt)
 	return QDateTime::fromTime_t(dt.toTime_t());
 }
 
+bool FileOpsTest::forceNodeTimeout()
+{
+	auto reply = m_kiofuseprivate_iface.forceNodeTimeout();
+	reply.waitForFinished();
+	return !reply.isError();
+}
+
 QTEST_GUILESS_MAIN(FileOpsTest)
 
 #include "fileopstest.moc"
